import cv2
import os
import re
import subprocess

from phycam_config import ar0144_formats, ar0234_formats, ar0521_formats
from phycam_config import ar0144_modes, ar0234_modes, ar0521_modes


class PhyCam():
    def __init__(self, port=0):

        self._modes = {
            "ar0144" : ar0144_modes,
            "ar0234" : ar0234_modes,
            "ar0521" : ar0521_modes,
        }
        self._formats = {
            "ar0144" : ar0144_formats,
            "ar0234" : ar0234_formats,
            "ar0521" : ar0521_formats,
        }
        self._names = {
            "ar0144" : "VM-016 / VM-116",
            "ar0234" : "VM-020 / VM-120",
            "ar0521" : "VM-017 / VM-117",
        }

        self._port = port

        if os.path.islink("/dev/cam-csi1"):
            self._subdev = "/dev/cam-csi1"
        elif os.path.islink("/dev/cam-csi1-port0") and port == 0:
            self._subdev = "/dev/cam-csi1-port0"
        elif os.path.islink("/dev/cam-csi1-port1") and port == 1:
            self._subdev = "/dev/cam-csi1-port1"
        else:
            raise RuntimeError("No phyCAM subdevice detected!")

        if os.path.islink("/dev/video-csi1"):
            self._videodev = "/dev/video-csi1"
        else:
            raise RuntimeError("No phyCAM videodevice detected!")


        real_subdev = os.path.realpath(self._subdev)
        name_path = f"/sys/class/video4linux/{real_subdev.split('/')[-1]}/name"

        with open(name_path, "r") as f:
            self._entity = f.read().splitlines()[0]

        self._driver = self._entity.split()[0]

        if self._driver not in self._modes:
            raise RuntimeError(f"Detected camera of type {self._driver} not supported!")

        command = ["v4l2-ctl", "-d", self._subdev, "--get-subdev-fmt", "0"]
        result = subprocess.run(command, capture_output=True)
        output = result.stdout

        color = re.search(br"MEDIA_BUS_FMT_([A-Z]+)\d+.*\)", output).group(1)
        self._color = color.decode("utf-8")

        self._name = self._names[self._driver]
        self._mode = self._modes[self._driver]["Full"]
        if self._color == "Y":
            self._format = self._formats[self._driver]["monochrome"]
        else:
            self._format = self._formats[self._driver]["color"]

        color = "BW" if self._color == "Y" else "COL"

        if self._subdev == "/dev/cam-csi1":
            interface = "CSI1"
        elif self._subdev == "/dev/cam-csi1-port0":
            interface = "CSI1 via phyCAM-L Port 0"
        else:
            interface = "CSI1 via phyCAM-L Port 1"

        print(f"Detected a {self._name} ({self._driver}) {color} on {interface}")


    @property
    def subdev(self):
        return self._subdev


    @property
    def videodev(self):
        return self._videodev

    @property
    def gst_pipeline(self):
        fmt = self._format
        mode = self._mode

        pipeline = [
            f"gst-launch-1.0 v4l2src device={self._videodev}",
            f"{fmt['gst']},width={mode['cap_width']},height={mode['cap_height']}",
            f"appsink",
        ]

        return " ! ".join(pipeline)

    @property
    def opencv_conv(self):
        return self._format["opencv"]


    def select_mode(self, mode):
        if mode in self._modes[self._driver]:
            self._mode = self._modes[self._driver][mode]
        else:
            raise ValueError(f"Mode {mode} not avialable for {self._driver}")


    def setup_pipeline(self):
        mode = self._mode
        fmt = self._format

        command = [
            "setup-pipeline-csi1",
            "-f", fmt["cam"],
            "-s", f"{mode['cap_width']}x{mode['cap_height']}",
            "-o", mode["offset"],
            "-c", f"{mode['width']}x{mode['height']}",
            "-p", str(self._port),
        ]

        subprocess.run(command)


    def set_controls(self):
        print("Setting V4L2 Controls:")
        print("----------------------")

        for control in self._format["controls"]:
            print(f"  v4l2-ctl -d {self._subdev} -c {control}")

        command = [
            "v4l2-ctl",
            "-d", f"{self._subdev}",
            "-c", ",".join(self._format["controls"])
        ]

        subprocess.run(command)

        for control in self._mode["controls"]:
            print(f"  v4l2-ctl -d {self._subdev} -c {control}")

        print("")

        command = [
            "v4l2-ctl",
            "-d", f"{self._subdev}",
            "-c", ",".join(self._mode["controls"])
        ]

        subprocess.run(command)

        if self._driver == "ar0521":
            print((
                f"  Since the {self._name} does not feature an internal AEC algorithm,\n"
                 "  using this example in low light conditions can result in very dark images.\n"
                 "  To improve this, you can increase the analogue gain setting with\n"
                f"     v4l2-ctl -d {self._subdev} -c analogue_gain=8000\n"
                 ""
            ))
